#
# SpatCompFuncs.R
#
#
# This file contains the MCMC functions and data generating functions for the 
# regualar array examples. 
#
#################################################################################
# Necessary Packages
rm(list= ls())

library(MASS)
library(Matrix) 
library(mvtnorm)
library(mnormt)



################################################################################
################################################################################
################################################################################
################################################################################
# 
# New block Gibbs sampler:
# 
#

new.block.gibbs<-function(Y,D,W,rho,id1,id2,id3,id4,phi,sig2,tau2,as2,bs2,at2,
                          bt2,G,verbose=FALSE){
  # This function contains the chromatic Gibbs sampler for a regular array, 
  # using four colorings
  
  V<-length(Y)
  DW<-D-rho*W                  
  D<-diag(D)
  
  n.id1<-length(id1)
  n.id2<-length(id2)
  n.id3<-length(id3)
  n.id4<-length(id4)
  
  
  #phi.save<-matrix(-99,nrow=G,ncol=V)
  #sig2.save<-rep(-99,G)
  #tau2.save<-rep(-99,G)
  #beta0.save<-rep(-99,G)
  phi.save<-0  # Time w/o saving results
  sig2.save<-0
  tau2.save<-0
  beta0.save<-0
  
  for(g in 1:G){
    
    ##############################################
    # Sample the intercept
    
    Y.s<-Y-phi
    
    beta0<-rnorm(1,mean(Y.s),sqrt(sig2/V))
    
    ##############################################
    # Block 1 update
    
    Yt<-Y-beta0
    
    #mui<-rho*W[id1,]%*%phi/D[id1]
    mui<-rho*W%*%phi/D  
    t2i<- tau2/D[id1]
    #pm<-as.vector((t2i*Yt[id1]+mui)/(1+t2i))
    pm<-as.vector((t2i*Yt[id1]+sig2*mui[id1])/(sig2+t2i))
    #ps<-as.vector(sqrt((sig2*t2i)/(1+t2i)))
    ps<-as.vector(sqrt((sig2*t2i)/(sig2+t2i)))
    phi1<-rnorm(n.id1,pm,ps)
    phi[id1]<-phi1
    
    ##############################################
    # Block 2 update
    
    #mui<-rho*W[id2,]%*%phi/D[id2]
    mui<-rho*W%*%phi/D
    t2i<- tau2/D[id2]
    #pm<-as.vector((t2i*Yt[id2]+mui)/(1+t2i))
    pm<-as.vector((t2i*Yt[id2]+sig2*mui[id2])/(sig2+t2i))
    #ps<-as.vector(sqrt((sig2*t2i)/(1+t2i)))
    ps<-as.vector(sqrt((sig2*t2i)/(sig2+t2i)))
    phi2<-rnorm(n.id2,pm,ps)
    phi[id2]<-phi2
    
    ##############################################
    # Block 3 update
    
    #mui<-rho*W[id3,]%*%phi/D[id3]
    mui<-rho*W%*%phi/D  
    t2i<- tau2/D[id3]
    #pm<-as.vector((t2i*Yt[id3]+mui)/(1+t2i))
    pm<-as.vector((t2i*Yt[id3]+sig2*mui[id3])/(sig2+t2i))
    #ps<-as.vector(sqrt((sig2*t2i)/(1+t2i)))
    ps<-as.vector(sqrt((sig2*t2i)/(sig2+t2i)))
    phi3<-rnorm(n.id3,pm,ps)
    phi[id3]<-phi3
    
    ##############################################
    # Block 4 update
    
    #mui<-rho*W[id4,]%*%phi/D[id4]
    mui<-rho*W%*%phi/D
    t2i<- tau2/D[id4]
    #pm<-as.vector((t2i*Yt[id4]+mui)/(1+t2i))
    pm<-as.vector((t2i*Yt[id4]+sig2*mui[id4])/(sig2+t2i))
    #ps<-as.vector(sqrt((sig2*t2i)/(1+t2i)))
    ps<-as.vector(sqrt((sig2*t2i)/(sig2+t2i)))
    phi4<-rnorm(n.id4,pm,ps)
    phi[id4]<-phi4
    
    # Center to improve identifiability
    phi <- phi - mean(phi)
    
    ###############################################
    # Sample sig2
    
    #as2s<-as2+(V+V)/2
    as2s<-as2 + V/2  # Oops!
    #bs2s<-as.vector(bs2+ sum((Y-beta0-phi)^2)/2 + phi%*%DW%*%phi/(2*tau2))
    bs2s<-as.vector(bs2+ sum((Y-beta0-phi)^2)/2)  # Oops!
    
    samp<-rgamma(1,as2s,bs2s)
    
    sig2<-1/samp
    
    ###############################################
    # Sample tau2
    
    #at2s<-at2+V/2
    at2s<-at2+(V-1)/2
    #bt2s<-as.vector(bt2+ phi%*%DW%*%phi/(2*sig2))
    bt2s<-as.vector(bt2+ t(phi)%*%DW%*%phi/(2))
    
    samp<-rgamma(1,at2s,bt2s)
    
    tau2<-1/samp
    
    #sig2.save[g]<-sig2
    #tau2.save[g]<-tau2
    #beta0.save[g]<-beta0
    #phi.save[g,]<-phi
    
    if(verbose==TRUE){
      print(g)
      
      if(g%%100==0){   
        par(mfrow=c(2,1))
        plot(sig2.save[1:g])
        plot(tau2.save[1:g])
      }
    } 
  }
  
  return(list("sig2"=sig2.save,"tau2"=tau2.save,"beta0"=beta0.save,"phi"=phi.save))
}



################################################################################
################################################################################
################################################################################

spam.gibbs <- function(Y,D,W,rho,phi,sig2,tau2,as2,bs2,at2,bt2,G,
                       verbose=FALSE){
  
  # This function implements the Gibbs sampler on the regular array example 
  # using block Gibbs updates. It uses the spam package, with options set to 
  # speed it up.
  
  
  require(spam)
  
  options(spam.cholsymmetrycheck= FALSE, spam.safemodevalidity= FALSE)
  
  
  
  W <- as.spam(as.matrix(W))
  
  V <- length(Y)
  DW <- diag.spam(diag(D)) - rho*W
  
  I <- diag.spam(rep(1,times= V))
  
  Rs <- chol(DW + I)  # This saves the sparsity structure
  
  # Get Cholesky factorization of Dw + noise covariance matrix
  
  # MCMC output storage
  phi.save<-matrix(-99,nrow=G,ncol=V)
  sig2.save<-rep(-99,G)
  tau2.save<-rep(-99,G)
  beta0.save<-rep(-99,G)
  
  # Center phi before beginning
  phi <- phi - mean(phi)
  
  for (g in 1:G){
    
    # Sample the intercept
    Y.s<-Y-phi
    
    beta0<-rnorm(1,mean(Y.s),sqrt(sig2/V))
    
    
    # Update the random effects
    Yt <- (Y-beta0)/sig2
    
    # Set \eta = sig2/tau2 for convenience
    eta <- tau2/sig2  
    
    Prec<- (1/sig2)*(I + DW/eta)
    
    phi <- c(rmvnorm.canonical(1, Yt, Prec, Rstruct= Rs))
    
    # Center for identifiability
    phi <- phi - mean(phi)
    
    
    
    
    ###############################################
    # Sample sig2
    
    as2s<-as2+V/2
    bs2s<-as.vector(bs2+ sum((Y-beta0-phi)^2)/2)
    
    
    samp<-rgamma(1,as2s,bs2s)
    
    sig2<-1/samp
    
    ###############################################
    # Sample tau2
    
    at2s<-at2+(V-1)/2
    #bt2s<-as.vector(bt2+ phi%*%DW%*%phi/2)
    bt2s<- as.vector(bt2+ t(phi)%*%DW%*%phi/2)
    
    samp<-rgamma(1,at2s,bt2s)
    
    tau2<-1/samp
    
    
    # Save output
    sig2.save[g]<-sig2
    tau2.save[g]<-tau2
    beta0.save[g]<-beta0
    phi.save[g,]<-phi
    
    if(verbose==TRUE){
      print(g)
      
      if(g%%100==0){   
        par(mfrow=c(2,1))
        plot(sig2.save[1:g])
        plot(tau2.save[1:g])
      }  # End g if
    }  # End verbose if
    
    
  }  # End MCMC iterations
  
  detach(package:spam, unload= TRUE)
  return(list("sig2"=sig2.save,"tau2"=tau2.save,"beta0"=beta0.save,"phi"=phi.save))
  
  
}



################################################################################
################################################################################
################################################################################
################################################################################

ind.gibbs<-function(Y,D,W,rho,phi,sig2,tau2,as2,bs2,at2,bt2,G,verbose=FALSE){
  # This function implements Gibbs sampling over the GMRF via single-site 
  # updates
  
  V<-length(Y)
  DW<-D-rho*W                  
  D<-diag(D)
  
  phi.save<-matrix(-99,nrow=G,ncol=V)
  sig2.save<-rep(-99,G)
  tau2.save<-rep(-99,G)
  beta0.save<-rep(-99,G)
  
  
  # Center phi before beginning
  phi <- phi - mean(phi)
  
  for(g in 1:G){
    
    ##############################################
    # Sample the intercept
    
    Y.s<-Y-phi
    
    beta0<-rnorm(1,mean(Y.s),sqrt(sig2/V))
    
    ##############################################
    # sample spatial random effects 
    
    Yt<-Y-beta0
    
    for(i in 1:V){
      #mui<-rho*W[i,]%*%phi/D[i]
      mui<-rho*sum(W[i,]*phi)/D[i]
      t2i<- tau2/D[i]
      pm<-as.vector((t2i*Yt[i] + sig2*mui)/(sig2 + t2i))
      ps<-as.vector(sqrt((sig2*t2i)/(sig2 + t2i)))
      phii<-rnorm(1,pm,ps)
      phi[i]<-phii
    }
    
    
    # Since we have an intercept in the model, re-center \phi at each iteration
    # for identifiability
    phi <- phi - mean(phi)
    
    
    ###############################################
    # Sample sig2
    
    #as2s<-as2+(V-1)/2
    as2s<-as2+V/2
    bs2s<-as.vector(bs2 + sum((Y-beta0-phi)^2)/2)
    
    samp<-rgamma(1,as2s,bs2s)
    
    sig2<-1/samp
    
    ###############################################
    # Sample tau2
    
    at2s<-at2+(V-1)/2
    #bt2s<-as.vector(bt2 + phi%*%DW%*%phi/2)
    bt2s<- as.vector(bt2 + t(phi)%*%DW%*%phi/2)
    
    samp<-rgamma(1,at2s,bt2s)
    
    tau2<-1/samp
    
    sig2.save[g]<-sig2
    tau2.save[g]<-tau2
    beta0.save[g]<-beta0
    phi.save[g,]<-phi
    
    if(verbose==TRUE){
      print(g)
      
      if(g%%100==0){   
        par(mfrow=c(2,1))
        plot(sig2.save[1:g])
        plot(tau2.save[1:g])
      }
    } 
  }
  
  return(list("sig2"=sig2.save,"tau2"=tau2.save,"beta0"=beta0.save,"phi"=phi.save))
}

ind.Rcpp.gibbs<-function(Y,D,W,rho,phi,sig2,tau2,as2,bs2,at2,bt2,G,verbose=FALSE){
  # This function implements Gibbs sampling over the GMRF via single-site 
  # updates in Rcpp
  
  V<-length(Y)
  DW<-D-rho*W                  
  D<-diag(D)
  #W.indices is an matrix containing the nonzero indices of W to be passed to SingleSite.cpp
  W.indices = matrix(NA,V,max(D))
  for(i in 1:V){
    col.ind = which(W[i,]==1)
    W.indices[i,1:D[i]] = col.ind
  }
  
  phi.save<-matrix(-99,nrow=G,ncol=V)
  sig2.save<-rep(-99,G)
  tau2.save<-rep(-99,G)
  beta0.save<-rep(-99,G)
  
  
  # Center phi before beginning
  phi <- phi - mean(phi)
  
  for(g in 1:G){
    
    ##############################################
    # Sample the intercept
    
    Y.s<-Y-phi
    
    beta0<-rnorm(1,mean(Y.s),sqrt(sig2/V))
    
    ##############################################
    # sample spatial random effects 
    
    Yt<-Y-beta0
    
    phi = SingleSite(rho, W.indices, phi, D, V, tau2, Yt, sig2)
    
    
    # Since we have an intercept in the model, re-center \phi at each iteration
    # for identifiability
    phi <- phi - mean(phi)
    
    
    ###############################################
    # Sample sig2
    
    #as2s<-as2+(V-1)/2
    as2s<-as2+V/2
    bs2s<-as.vector(bs2 + sum((Y-beta0-phi)^2)/2)
    
    samp<-rgamma(1,as2s,bs2s)
    
    sig2<-1/samp
    
    ###############################################
    # Sample tau2
    
    at2s<-at2+(V-1)/2
    #bt2s<-as.vector(bt2 + phi%*%DW%*%phi/2)
    bt2s<- as.vector(bt2 + t(phi)%*%DW%*%phi/2)
    
    samp<-rgamma(1,at2s,bt2s)
    
    tau2<-1/samp
    
    sig2.save[g]<-sig2
    tau2.save[g]<-tau2
    beta0.save[g]<-beta0
    phi.save[g,]<-phi
    
    if(verbose==TRUE){
      print(g)
      
      if(g%%100==0){   
        par(mfrow=c(2,1))
        plot(sig2.save[1:g])
        plot(tau2.save[1:g])
      }
    } 
  }
  
  return(list("sig2"=sig2.save,"tau2"=tau2.save,"beta0"=beta0.save,"phi"=phi.save))
}

parallel.chromatic.gibbs<-function(Y,D,W,rho,id1,id2,id3,id4,phi,sig2,tau2,as2,bs2,at2,
                          bt2,G,verbose=FALSE){

  # This function contains the chromatic Gibbs sampler for a regular array,
  # using four colorings, and parallel updates within a chromatic block.  This function requires 
  # that all blocks have the same number of nodes.
  
  V<-length(Y)
  DW<-D-rho*W                  
  D<-diag(D)
  
  n.id1<-length(id1)
  n.id2<-length(id2)
  n.id3<-length(id3)
  n.id4<-length(id4)
  
  
  #phi.save<-matrix(-99,nrow=G,ncol=V)
  #sig2.save<-rep(-99,G)
  #tau2.save<-rep(-99,G)
  #beta0.save<-rep(-99,G)
  phi.save<-0  # For when we want to time it, but don't
    # want to save the results so we can preserve memory
  sig2.save<-0
  tau2.save<-0
  beta0.save<-0
  
  
  # # Determine the number of cores to parallelize over.  The number of cores must evenly divide the
  # # number of nodes in each block.
  # pfv = sort(unique(primeFactors(V/4)))
  # nocores = min(pfv)
  # if(nocores==2){
  #   if(length(pfv)>1){
  #     nocores = pfv[2]
  #   }else{
  #     nocores = 4
  #   }
  # }
  # cl = makeCluster(nocores)
  
  for(g in 1:G){
    
    ##############################################
    # Sample the intercept
    
    Y.s<-Y-phi
    
    beta0<-rnorm(1,mean(Y.s),sqrt(sig2/V))
    
    ##############################################
    # Block 1 update
    
    # Yt<-Y-beta0
    # 
    # #mui<-rho*W[id1,]%*%phi/D[id1]
    # mui<-rho*W%*%phi/D  
    # t2i<- tau2/D[id1]
    # #pm<-as.vector((t2i*Yt[id1]+mui)/(1+t2i))
    # pm<-as.vector((t2i*Yt[id1]+sig2*mui[id1])/(sig2+t2i))
    # #ps<-as.vector(sqrt((sig2*t2i)/(1+t2i)))
    # ps<-as.vector(sqrt((sig2*t2i)/(sig2+t2i)))
    # phi1<-rnorm(n.id1,pm,ps)
    # phi[id1]<-phi1
    
    Yt<-Y-beta0
    
    #mui<-rho*W[id1,]%*%phi/D[id1]
    mui<-rho*W%*%phi/D  
    t2i<- tau2/D[id1]
    #pm<-as.vector((t2i*Yt[id1]+mui)/(1+t2i))
    pm<-as.vector((t2i*Yt[id1]+sig2*mui[id1])/(sig2+t2i))
    #ps<-as.vector(sqrt((sig2*t2i)/(1+t2i)))
    ps<-as.vector(sqrt((sig2*t2i)/(sig2+t2i)))
    ###Divide the vector of standard deviations into a piece for each core
    x.list = list()
    for(i in 1:nocores){
      x.list[[i]] = ps[((i-1)*length(ps)/nocores +1):(i*length(ps)/nocores)]
    }
    #Sample the phi's in parallel.  To prevent having to pass a large mean vector to each core, the
    #phi's are sampled with mean 0, and then the mean is added to the result are the samples from each 
    #core are have been recombined.
    phi1 = clusterApply(cl,x.list,rnorm,n=length(ps)/nocores,mean = 0)
    phi[id1]<-unlist(phi1) + pm
    
    ##############################################
    # Block 2 update
    
    #mui<-rho*W[id2,]%*%phi/D[id2]
    mui<-rho*W%*%phi/D
    t2i<- tau2/D[id2]
    #pm<-as.vector((t2i*Yt[id2]+mui)/(1+t2i))
    pm<-as.vector((t2i*Yt[id2]+sig2*mui[id2])/(sig2+t2i))
    #ps<-as.vector(sqrt((sig2*t2i)/(1+t2i)))
    ps<-as.vector(sqrt((sig2*t2i)/(sig2+t2i)))
    x.list = list()
    for(i in 1:nocores){
      x.list[[i]] = ps[((i-1)*length(ps)/nocores +1):(i*length(ps)/nocores)]
    }
    phi2 = clusterApply(cl,x.list,rnorm,n=length(ps)/nocores,mean = 0)
    phi[id2]<-unlist(phi2) + pm
    
    ##############################################
    # Block 3 update
    
    #mui<-rho*W[id3,]%*%phi/D[id3]
    mui<-rho*W%*%phi/D
    t2i<- tau2/D[id3]
    #pm<-as.vector((t2i*Yt[id3]+mui)/(1+t2i))
    pm<-as.vector((t2i*Yt[id3]+sig2*mui[id3])/(sig2+t2i))
    #ps<-as.vector(sqrt((sig2*t2i)/(1+t2i)))
    ps<-as.vector(sqrt((sig2*t2i)/(sig2+t2i)))
    x.list = list()
    for(i in 1:nocores){
      x.list[[i]] = ps[((i-1)*length(ps)/nocores +1):(i*length(ps)/nocores)]
    }
    phi3 = clusterApply(cl,x.list,rnorm,n=length(ps)/nocores,mean = 0)
    phi[id3]<-unlist(phi3) + pm
    
    ##############################################
    # Block 4 update
    
    #mui<-rho*W[id4,]%*%phi/D[id4]
    mui<-rho*W%*%phi/D  
    t2i<- tau2/D[id4]
    #pm<-as.vector((t2i*Yt[id4]+mui)/(1+t2i))
    pm<-as.vector((t2i*Yt[id4]+sig2*mui[id4])/(sig2+t2i))
    #ps<-as.vector(sqrt((sig2*t2i)/(1+t2i)))
    ps<-as.vector(sqrt((sig2*t2i)/(sig2+t2i)))
    x.list = list()
    for(i in 1:nocores){
      x.list[[i]] = ps[((i-1)*length(ps)/nocores +1):(i*length(ps)/nocores)]
    }
    phi4 = clusterApply(cl,x.list,rnorm,n=length(ps)/nocores,mean = 0)
    phi[id4]<-unlist(phi4) + pm
    
    # Center to improve identifiability
    phi <- phi - mean(phi)
    
    ###############################################
    # Sample sig2
    
    #as2s<-as2+(V+V)/2
    as2s<-as2+V/2
    #bs2s<-as.vector(bs2+ sum((Y-beta0-phi)^2)/2 + phi%*%DW%*%phi/(2*tau2))
    bs2s<-as.vector(bs2+ sum((Y-beta0-phi)^2)/2)
    
    samp<-rgamma(1,as2s,bs2s)
    
    sig2<-1/samp
    
    ###############################################
    # Sample tau2
    
    #at2s<-at2+V/2
    at2s<-at2+(V-1)/2
    #bt2s<-as.vector(bt2+ phi%*%DW%*%phi/(2*sig2))
    bt2s<-as.vector(bt2+ phi%*%DW%*%phi/2)
    
    samp<-rgamma(1,at2s,bt2s)
    
    tau2<-1/samp
    
    #sig2.save[g]<-sig2
    #tau2.save[g]<-tau2
    #beta0.save[g]<-beta0
    #phi.save[g,]<-phi
    
    if(verbose==TRUE){
      print(g)
      
      if(g%%100==0){   
        par(mfrow=c(2,1))
        plot(sig2.save[1:g])
        plot(tau2.save[1:g])
      }
    } 
  }
  #stopCluster(cl)
  
  #sp = Sys.time()
  #time5[1] = sp-st
  return(list("sig2"=sig2.save,"tau2"=tau2.save,"beta0"=beta0.save,"phi"=phi.save))
}


################################################################################
################################################################################
################################################################################
################################################################################

data.gen<-function(P1,P2,sig2=1){
  # This function simulates a P1XP2 regular array of data using the scaled
  # Gaussian density as the true image, and corrupting it with Gaussian noise.
  # It also returns the vectors identifying each graph color.
  
  #P1<-50
  #P2<-50
  V<-P1*P2
  Vid<-array(1:V,c(P1,P2)) # Arbitrarily identifies voxels   
  
  Y<-rep(-99,V) # Arbitrary response vector                 
  Dj<-rep(-99,V)
  Nj<-matrix(-99,nrow=V,ncol=8)  # This will list the neighbors for each voxel
  
  id1<-rep(-99,V)  # Storing the sets of indices for each independent group
  id2 <- rep(-99,V)
  id3<-rep(-99,V)
  id4<-rep(-99,V)
  
  c1<-1
  c2<-1
  c3<-1
  c4<-1
  
  for(p1 in 1:P1){
    for(p2 in 1:P2){
      
      # Create the neighborhood matrix, the design matrix, and the segmentation
      # vectors
      l1<-max(c(1,p1-1))
      l2<-max(c(1,p2-1))
      u1<-min(c(P1,p1+1))
      u2<-min(c(P2,p2+1))
      
      neighbors<-as.vector(Vid[l1:u1,l2:u2])
      neighbors<-neighbors[neighbors!=Vid[p1,p2]]
      
      # Y is \propto the ratio of a MVN(c(25,25), diag(25,25)) density to the MVN
      # evaluated at the mean + noise (assuming P1 = P2 = 25)
      Y[Vid[p1,p2]]<- 10*exp(dmnorm(c(p1,p2),c(P1/2,P2/2),diag(c(P1/2,P2/2)),
                                    log=TRUE) -
                               dmnorm(c(P1/2,P2/2),c(P1/2,P2/2),
                                      diag(c(P1/2,P2/2)),log=TRUE)) + 
        rnorm(1,0,sqrt(sig2))
      
      wjv<-length(neighbors)
      Dj[Vid[p1,p2]]      <-wjv
      Nj[Vid[p1,p2],1:wjv]<-neighbors 
      
      # Assign voxel to conditionally idependent group based on location
      if(p1%%2!=0 & p2%%2!=0){ 
        id1[c1]<-Vid[p1,p2]; c1<-c1+1
      }
      if(p1%%2!=0 & p2%%2==0){ 
        id2[c2]<-Vid[p1,p2]; c2<-c2+1
      }
      if(p1%%2==0 & p2%%2!=0){ 
        id3[c3]<-Vid[p1,p2]; c3<-c3+1
      }
      if(p1%%2==0 & p2%%2==0){ 
        id4[c4]<-Vid[p1,p2]; c4<-c4+1
      }
      
      
    }
  }
  
  id1<-id1[id1!=-99]
  id2<-id2[id2!=-99]
  id3<-id3[id3!=-99]
  id4<-id4[id4!=-99]
  
  n.id1<-length(id1)
  n.id2<-length(id2)
  n.id3<-length(id3)
  n.id4<-length(id4)
  
  ######################################################
  # Creating the necessary sparse matrices
  i<-NULL
  j<-NULL
  
  for(v in 1:V){
    i<-c(i,rep(v,Dj[v]))
    j<-c(j,Nj[v,1:Dj[v]])
  }
  
  W<- sparseMatrix(i, j, x = 1)  
  D<-sparseMatrix(1:V,1:V,x=Dj)
  
  return(list("Y"=Y,"D"=D,"W"=W,"id1"=id1,"id2"=id2,"id3"=id3,"id4"=id4))
}










